from decimal import Decimal

import pytest

from seqscore.encoding import EncodingError
from seqscore.model import LabeledSequence, Mention, SequenceProvenance, Span
from seqscore.scoring import (
    AccuracyScore,
    ClassificationScore,
    TokenCountError,
    compute_scores,
    convert_score,
    score_label_sequences,
    score_sequence_label_accuracy,
    score_sequence_mentions,
)


def test_score_sentence_labels_correct() -> None:
    ref_labels = ["O", "B-ORG", "I-ORG", "O"]
    pred_labels = ref_labels[:]
    score = AccuracyScore()
    score_sequence_label_accuracy(pred_labels, ref_labels, score)
    assert score.total == 4
    assert score.hits == 4
    assert score.accuracy == 1.0


def test_score_sentence_labels_incorrect() -> None:
    ref_labels = ["O", "B-ORG", "I-ORG", "O"]
    pred_labels = ref_labels[:]
    pred_labels[2] = "B-LOC"
    score = AccuracyScore()
    score_sequence_label_accuracy(pred_labels, ref_labels, score)
    assert score.total == 4
    assert score.hits == 3
    assert score.accuracy == pytest.approx(3 / 4)


def test_score_sentence_labels_invalid() -> None:
    ref_labels = ["O", "B-ORG", "I-ORG", "O"]
    # Shorter predictions than reference
    pred_labels = ref_labels[:-1]
    with pytest.raises(ValueError):
        score_sequence_label_accuracy(pred_labels, ref_labels, AccuracyScore())


def test_score_sentence_mentions_correct() -> None:
    ref_mentions = [Mention(Span(0, 2), "PER"), Mention(Span(4, 5), "ORG")]
    pred_mentions = [Mention(Span(0, 2), "PER"), Mention(Span(4, 5), "ORG")]
    score = ClassificationScore()
    score_sequence_mentions(pred_mentions, ref_mentions, score)
    assert score.true_pos == 2
    assert score.false_pos == 0
    assert score.false_neg == 0
    assert score.type_scores == {
        "PER": ClassificationScore(true_pos=1),
        "ORG": ClassificationScore(true_pos=1),
    }
    assert score.total_ref == 2
    assert score.total_pos == 2
    assert score.precision == 1.0
    assert score.recall == 1.0
    assert score.f1 == 1.0


def test_score_sentence_mentions_incorrect1() -> None:
    ref_mentions = [
        Mention(Span(0, 2), "LOC"),
        Mention(Span(4, 5), "PER"),
        Mention(Span(7, 8), "MISC"),
        Mention(Span(9, 11), "MISC"),
    ]
    pred_mentions = [
        Mention(Span(0, 2), "ORG"),
        Mention(Span(4, 5), "PER"),
        Mention(Span(9, 11), "MISC"),
    ]
    score = ClassificationScore()
    score_sequence_mentions(pred_mentions, ref_mentions, score)
    assert score.true_pos == 2
    assert score.false_pos == 1
    assert score.false_neg == 2
    assert score.type_scores == {
        "PER": ClassificationScore(true_pos=1),
        "LOC": ClassificationScore(false_neg=1),
        "MISC": ClassificationScore(false_neg=1, true_pos=1),
        "ORG": ClassificationScore(false_pos=1),
    }
    assert score.total_ref == 4
    assert score.total_pos == 3
    assert score.precision == pytest.approx(2 / 3)
    assert score.recall == pytest.approx(2 / 4)
    # Note that we have already checked the precision and recall values
    assert score.f1 == pytest.approx(
        2 * (score.precision * score.recall) / (score.precision + score.recall)
    )


def test_score_label_sequences_correct() -> None:
    ref_labels = [["O", "B-ORG", "I-ORG", "O"], ["B-PER", "I-PER"]]
    pred_labels = ref_labels[:]
    classification, accuracy = score_label_sequences(
        pred_labels, ref_labels, "BIO", repair=None
    )

    assert accuracy.total == 6
    assert accuracy.hits == 6
    assert accuracy.accuracy == 1.0

    assert classification.true_pos == 2
    assert classification.false_pos == 0
    assert classification.false_neg == 0
    assert classification.type_scores["ORG"] == ClassificationScore(true_pos=1)
    assert classification.type_scores["PER"] == ClassificationScore(true_pos=1)


def test_score_label_sequences_invalid_norepair() -> None:
    ref_labels = [["O", "B-ORG", "I-ORG", "O"], ["B-PER", "I-PER"]]
    pred_labels = [["O", "B-ORG", "I-ORG", "O"], ["I-PER", "I-PER"]]
    with pytest.raises(EncodingError):
        score_label_sequences(pred_labels, ref_labels, "BIO", repair=None)


def test_score_label_sequences_invalid_repair() -> None:
    ref_labels = [["O", "B-ORG", "I-ORG", "O"], ["B-PER", "I-PER"]]
    pred_labels = [["O", "I-ORG", "I-ORG", "O"], ["O", "I-PER"]]
    classification, accuracy = score_label_sequences(
        pred_labels, ref_labels, "BIO", repair="conlleval"
    )

    assert accuracy.total == 6
    assert accuracy.hits == 4
    assert accuracy.accuracy == 4 / 6

    assert classification.true_pos == 1
    assert classification.false_pos == 1
    assert classification.false_neg == 1
    assert classification.type_scores["ORG"] == ClassificationScore(true_pos=1)
    assert classification.type_scores["PER"] == ClassificationScore(
        false_pos=1, false_neg=1
    )


def test_classification_score_empty() -> None:
    score = ClassificationScore()
    assert score.precision == 0.0
    assert score.recall == 0.0
    assert score.f1 == 0.0


def test_classification_score_update() -> None:
    score1 = ClassificationScore()
    score1.true_pos += 1
    score1.type_scores["PER"].true_pos += 1
    score1.false_pos += 1
    score1.type_scores["ORG"].false_pos += 1

    score2 = ClassificationScore()
    score2.false_pos += 1
    score2.type_scores["ORG"].false_pos += 1
    score2.false_neg += 1
    score2.type_scores["MISC"].false_neg += 1
    score2.true_pos += 4
    score2.type_scores["ORG"].true_pos += 4

    score1.update(score2)

    assert score1.true_pos == 5
    assert score1.false_pos == 2
    assert score1.false_neg == 1
    assert score1.type_scores == {
        "PER": ClassificationScore(true_pos=1),
        "ORG": ClassificationScore(true_pos=4, false_pos=2),
        "MISC": ClassificationScore(false_neg=1),
    }


def test_accuracy_score_empty() -> None:
    score = AccuracyScore()
    assert score.accuracy == 0.0


def test_token_count_error() -> None:
    ref_labels = ["O", "B-ORG", "I-ORG", "O"]
    pred_labels = ["O", "B-ORG", "I-ORG", "O", "O"]
    ref_sequence = LabeledSequence(
        ["a", "b", "c", "d"], ref_labels, provenance=SequenceProvenance(0, "test")
    )
    pred_sequence = LabeledSequence(
        ["a", "b", "c", "d", "e"], pred_labels, provenance=SequenceProvenance(0, "test")
    )
    with pytest.raises(TokenCountError):
        compute_scores([[pred_sequence]], [[ref_sequence]])


def test_provenance_none_raises_error() -> None:
    labels = ["O", "B-ORG"]
    sequence = LabeledSequence(["a", "b"], labels, provenance=None)
    with pytest.raises(ValueError):
        TokenCountError.from_predicted_sequence(2, sequence)


def test_differing_num_docs() -> None:
    ref_labels = ["O", "B-ORG"]
    pred_labels = ["O", "B-LOC"]
    ref_sequence = LabeledSequence(
        ["a", "b"], ref_labels, provenance=SequenceProvenance(0, "test")
    )
    pred_sequence = LabeledSequence(
        ["a", "b"], pred_labels, provenance=SequenceProvenance(0, "test")
    )
    with pytest.raises(ValueError):
        compute_scores([[pred_sequence]], [[ref_sequence], [ref_sequence]])


def test_differing_doc_length() -> None:
    ref_labels = ["O", "B-ORG"]
    pred_labels = ["O", "B-LOC"]
    ref_sequence = LabeledSequence(
        ["a", "b"], ref_labels, provenance=SequenceProvenance(0, "test")
    )
    pred_sequence = LabeledSequence(
        ["a", "b"], pred_labels, provenance=SequenceProvenance(0, "test")
    )
    with pytest.raises(ValueError):
        compute_scores([[pred_sequence]], [[ref_sequence, ref_sequence]])


def test_differing_pred_and_ref_tokens() -> None:
    ref_labels = ["O", "B-ORG"]
    pred_labels = ["O", "B-LOC"]
    ref_sequence = LabeledSequence(
        ["a", "b"], ref_labels, provenance=SequenceProvenance(0, "test")
    )
    pred_sequence = LabeledSequence(
        ["a", "c"], pred_labels, provenance=SequenceProvenance(0, "test")
    )
    with pytest.raises(ValueError):
        compute_scores([[pred_sequence]], [[ref_sequence]])


def test_convert_score() -> None:
    # Check basic rounding up/down
    assert convert_score(0.92156, False) == Decimal("92.16")
    assert convert_score(0.92154, False) == Decimal("92.15")

    # Check half rounding
    # Note: due to inexact float representation, changing the test values
    # can lead to unexpected failures. If the final 5 is actually represented
    # as 49999 instead, it will cause rounding down.
    # See: https://docs.python.org/3/library/functions.html#round
    assert convert_score(0.03205, False) == Decimal("3.21")
    assert convert_score(0.03225, False) == Decimal("3.23")
    assert convert_score(0.02205, False) == Decimal("2.21")
    assert convert_score(0.02245, False) == Decimal("2.25")

    # Check that the number of decimal places is constant
    assert convert_score(1.0, False) == Decimal("100.00")
    assert convert_score(0.5, False) == Decimal("50.00")
    assert convert_score(0.0, False) == Decimal("0.00")

    # Check full precision
    assert convert_score(1 / 3, True) == 1 / 3
    assert convert_score(1 / 7, True) == 1 / 7
    assert convert_score(1 / 9, True) == 1 / 9
